/*** Yet Another Alpha Processor Simulator    ***/
/*** in C++    since 2002-05-15 by Kenji KISE ***/
/************************************************/
#include "define.h"

/************************************************/
void usage(){
  char UsageMessage[] =  "\
Usage: SimAlpha [-option] command\n\
 -d: debug mode\n\
 -c: print systemcall trace\n\
 -v[num]: verbose mode, prints '.'\
by VERBOSE_INTERVAL\n\
 -f[num]: set fd_adjust to verify simulation\n\
 -e[num][m]: end simulation at num code run. \
m means million\n\
 -i[file_name]: file_name as STDIN input file\n\
";
  printf("%s", UsageMessage); exit(0);
}

/************************************************/
int data_t::init(uint64_t d){
  value = d;
  cmov  = 0;
  return 0;
}

/************************************************/
uint64_t data_t::ld(){
  return value;
}

/************************************************/
int data_t::st(uint64_t d){
  value = d;
  return 0;
}

/*
 * Initialize the registers, data file format.
 * Begin by "\@" to indicate the valid data.
 * /@reg 28 0000000000000000
 * /@reg 29 0000000140023e90                    */
/************************************************/
architecture_state::architecture_state(){
  for(int i=0; i<32; i++) r[i].init(0);
  for(int i=0; i<32; i++) f[i].init(0);
  pc.init(0); 
}

uint64_t architecture_state::rr(int n){
  if(n==0) return pc.ld();
  if(n>=1  && n<=32) return r[n-1].ld();
  if(n>=33 && n<=64) return f[n-33].ld();
  printf("Error:architecture_state::rr(%d)\n",n);
  exit(1);
  return 0;
}

/************************************************/
void print_evaluation_result(struct evaluation_result *e,
			     struct system_config *ms, 
			     system_manager *sys){
  struct timeval tp;    
  struct timezone tzp;  
  gettimeofday(&tp, &tzp);
  
  uint64_t run_usec = (uint64_t)(tp.tv_sec - e->tp.tv_sec) * 1000000ull
    + (uint64_t)tp.tv_usec - (uint64_t)e->tp.tv_usec;
  fflush(stdout);
  fflush(stderr);
  int minute = (run_usec / 1000000) / 60;
  int second = (run_usec / 1000000) % 60;
  fprintf(stderr,"\n==================================");
  fprintf(stderr,"================================\n");
  fprintf(stderr,"=== %s\n", VER);
  char buf[256];
  getcwd(buf, 256);
  fprintf(stderr, "=== cwd: %s\n", buf);
  fprintf(stderr,"=== %5qd million code(%13qd code)", 
	  e->retired_inst/1000000, e->retired_inst);
  fprintf(stderr,"  %6.3f MIPS\n", 
	  (double)e->retired_inst/(double)run_usec);
  fprintf(stderr,"=== SimAlpha takes %3d min %2d second (%qd usec)\n", 
	  minute, second, run_usec);
  fprintf(stderr,"=== SimAlpha starts at %s", ctime(&e->time_begin));
  fprintf(stderr,"=== Memory %dKB used (block size %dKB, %d blocks)\n",
	  BLOCK_SIZE/1024 * e->used_memory_block,
	  BLOCK_SIZE/1024, e->used_memory_block);
  fflush(stderr);
}

/************************************************/
system_config::system_config(char *program, 
			     char **option){
  memset(this, 0, sizeof(system_config));
  strcpy(program_name, program);
  
  if(option==NULL) return;
  for(int i=0; option[i]!=NULL; i++){
    char opt[1024];
    strcpy(opt, option[i]);
    if(opt[0]!='-') continue;
    switch(opt[1]){
    case 'c': syscall_flag   = 1; break;
    case 'd': debug_flag     = 1; break;
    case 'v':
      verbose_interval =atoi(&opt[2]);
      if(verbose_interval == 0)   
	verbose_interval = 100000;
      if(opt[strlen(opt)-1]=='m') 
	verbose_interval*=1000000;
      printf("*** verbose interval: %d\n", 
	     verbose_interval);
      break;
    case 'f': 
      fd_adjust = atoi(&opt[2]);
      if(fd_adjust==0) fd_adjust=1;
      printf("*** fd_adjust=%d \n",fd_adjust);
      if(fd_adjust<0){
	printf("Error fd_adjust.\n");
	exit(1);
      }
      break;
    case 'e':
      end_tc =atoi(&opt[2]);
      if(opt[strlen(opt)-1]=='m') end_tc*=1000000;
      printf("*** Finish at %qd code execution\n",
	     end_tc);
      break;
    case 'i':
      fd_stdin_flag = 1;
      fd_stdin = open(&opt[2], O_RDONLY);
      printf("*** SimAlpha: use [%s] as STDIN.\n", 
	     &opt[2]);
      if(fd_stdin<0){
	printf("Can not open file: %s\n", &opt[2]); 
	exit(1);
      }
      break;
    default:
      printf("*** Error: [%s] Wrong option!!\n\n", 
	     opt);
      usage();
      exit(1);
    }
  }
}
