/*** Yet Another Alpha Processor Simulator    ***/
/*** in C++    since 2002-05-15 by Kenji KISE ***/
/************************************************/
#include "define.h"
#include "loader.h"

#define VER "SimAlpha Loader ver 1.0.0 2003-07-07"

/************************************************/
void usage_loader(){
  char UsageMessage[] =  "\
Usage: SimAlpha-Loader executable args\n\
  COFF format alpha executable\n\
";
  printf("%s", UsageMessage); exit(0);
}

/************************************************/
void format_check(char *p){
  FILE *fp;
  struct DU_filehdr hdr;
  struct DU_aouthdr aout;
  struct DU_scnhdr  st;

  if ((fp = fopen(p, "r")) == NULL) {
    fprintf(stderr, "Bad file name: %s\n", p);
    exit(0);
  }
  fread(&hdr,  sizeof(hdr), 1, fp);
  fread(&aout, sizeof(aout),1, fp);
  
  fprintf(stderr, "hdr.f_magic: %d,\n", 
	  hdr.f_magic);
  if(hdr.f_magic!=387){ /* elf 17791 */
    printf("This is not coff executable: %s.\n",p);
    exit(0);
  }

  for(int i=0; i<hdr.f_nscns; i++){
    fread(&st, sizeof(st), 1, fp);
    fprintf(stderr, "name: %8s  %09qx - %09qx \n", 
	    st.s_name, st.s_paddr, 
	    st.s_paddr + st.s_size);
    if(!strncmp(st.s_name, ".dynamic", 8)){
      printf("Using the shared library.\n");
      printf("opt -static for gcc.\n");
      printf("opt -non_shared for Digital.\n");
      exit(0);
    }
  }
  fclose(fp);
}

/************************************************/
void init_section(char *p, memory_system *mem, 
		  architecture_state *as){
  struct DU_filehdr hdr;
  struct DU_aouthdr aout;
  struct DU_scnhdr  st;
  
  FILE *fp = fopen(p, "r");
  fread(&hdr, sizeof(hdr), 1, fp);
  fread(&aout,sizeof(aout),1, fp);

  for(int i=0; i<hdr.f_nscns; i++){
    fread(&st, sizeof(st), 1, fp);
    if(!strcmp(st.s_name, ".sbss") || 
       !strcmp(st.s_name, ".bss") || 
       !strcmp(st.s_name, ".comment")) continue;
    
    int size = st.s_size / 4;
    int *data = new int[size];
    {
      FILE *f = fopen(p, "r");
      fseek(f, st.s_scnptr, SEEK_SET);
      fread(data, st.s_size, 1, f);
      fclose(f);
    }
    for(int j=0; j<size; j++){
      data_t dat, adr;
      dat.st(data[j]);
      adr.st(st.s_paddr + 4*j);
      mem->st_nbyte(4, &adr, &dat); /* store */
    }
    delete data;
  }
  fclose(fp);
  as->pc.st(aout.entry); /* Program Counter */
}

/************************************************/
#define BASE_STACK 0x11ff97000ull
#define BASE_ARGV  0x11ff98000ull
#define BASE_ARGP  0x11ff99000ull
void init_stack(int argc, char **argv, 
		memory_system *mem, 
		architecture_state *as){
  data_t dat, adr;
  adr.st(BASE_STACK);
  as->r[30] = adr; /* stack pointer */

  int argpc = 1;
  static char *argp[] ={"HOME=/"};

  /***** Error Checck *****/
  if(argc>15 || argpc>15){
    printf("Error: argc %d, argpc %d\n", 
	   argc, argpc);
    exit(1);
  }
  /***** argc *****/
  dat.st(argc);
  mem->st_nbyte(8, &adr, &dat);
  /***** argv *****/
  for(int i=0; i<argc; i++){
    adr.st(adr.ld()+8);
    dat.st(BASE_ARGV + 0x100*i);
    mem->st_nbyte(8, &adr, &dat);

    char *c = argv[i];
    for(unsigned int j=0; j<strlen(c); j++){
      data_t d, a;
      a.st(dat.ld() + j);
      d.st(c[j]);
      mem->st_nbyte(1, &a, &d);
    }
  }
  adr.st(adr.ld()+8);
  /***** argp *****/
  for(int i=0; i<argpc; i++){
    adr.st(adr.ld()+8);
    dat.st(BASE_ARGP + 0x100*i);
    mem->st_nbyte(8, &adr, &dat); /* argp */

    char *c = argp[i];
    for(unsigned int j=0; j<strlen(c); j++){
      data_t d, a;
      a.st(dat.ld() + j);
      d.st(c[j]);
      mem->st_nbyte(1, &a, &d);
    }
  }
}

/************************************************/
void create_image(architecture_state *as,
		  memory_system *mem){
  printf("/* SimAlpha 1.0 Image File */\n");

  printf("/*** Registers ***/\n");
  for(int i=0; i<32; i++)
    if(as->r[i].ld())
      printf("/@reg %02d %016qx\n", 
	     i, as->r[i].ld());
  printf("/@pc  32 %016qx\n", as->pc.ld());
  
  printf("/*** Memory    ***/\n");
  for(int i=0; i<BLOCK_TABLE_SIZE; i++){
    if(mem->mm->block_table[i]!=NULL){
      for(int j=0; j<BLOCK_SIZE/8; j++){
	data_t dat, adr;
	adr.st(0x100000000ull + (i<<13) + j*8);
	mem->ld_nbyte(8, &adr, &dat);
	if(dat.ld()!=0){
	  printf("@%qx %qx\n", 
		 adr.ld(), dat.ld());
	}
      }	
    }
  }
}

/************************************************/
int main(int argc, char **argv){
  fprintf(stderr, "%s\n", VER); 
  if(argc==1) usage_loader();
  char *p = argv[1]; /* program name */
  architecture_state as;
  memory_system mem;

  format_check(p);
  init_section(p, &mem, &as);
  init_stack(argc-1, &argv[1], &mem, &as);
  create_image(&as, &mem);
  return 0;
}
/************************************************/
